@@ -20,6 +20,7 @@ import contextlib
 from distutils import version
 import logging
 import string
+import threading
 
 from concurrent import futures
 from oslo_utils import encodeutils
@@ -45,19 +46,22 @@ def _translate_failures():
     try:
         yield
     except (exceptions.ConnectionError, exceptions.TimeoutError) as e:
-        coordination.raise_with_cause(coordination.ToozConnectionError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        utils.raise_with_cause(coordination.ToozConnectionError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
     except exceptions.RedisError as e:
-        coordination.raise_with_cause(coordination.ToozError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        utils.raise_with_cause(tooz.ToozError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
 
 
 class RedisLock(locking.Lock):
     def __init__(self, coord, client, name, timeout):
         name = "%s_%s_lock" % (coord.namespace, six.text_type(name))
         super(RedisLock, self).__init__(name)
+        # NOTE(jd) Make sure we don't release and heartbeat at the same time by
+        # using a exclusive access lock (LP#1557593)
+        self._exclusive_access = threading.Lock()
         self._lock = client.lock(name,
                                  timeout=timeout,
                                  thread_local=False)
@@ -82,32 +86,34 @@ class RedisLock(locking.Lock):
             acquired = self._lock.acquire(
                 blocking=blocking, blocking_timeout=timeout)
             if acquired:
-                self._coord._acquired_locks.add(self)
+                with self._exclusive_access:
+                    self._coord._acquired_locks.add(self)
             return acquired
 
     def release(self):
-        if not self.acquired:
-            return False
-        with _translate_failures():
-            try:
-                self._lock.release()
-            except exceptions.LockError:
+        with self._exclusive_access:
+            if not self.acquired:
                 return False
-            self._coord._acquired_locks.discard(self)
-            return True
+            with _translate_failures():
+                try:
+                    self._lock.release()
+                except exceptions.LockError:
+                    return False
+                self._coord._acquired_locks.discard(self)
+                return True
 
     def heartbeat(self):
-        if self.acquired:
-            with _translate_failures():
-                self._lock.extend(self._lock.timeout)
+        with self._exclusive_access:
+            if self.acquired:
+                with _translate_failures():
+                    self._lock.extend(self._lock.timeout)
 
     @property
     def acquired(self):
         return self in self._coord._acquired_locks
 
 
-class RedisDriver(coordination._RunWatchersMixin,
-                  coordination.CoordinationDriver):
+class RedisDriver(coordination.CoordinationDriverCachedRunWatchers):
     """Redis provides a few nice benefits that act as a poormans zookeeper.
 
     It **is** fully functional and implements all of the coordination
@@ -316,7 +322,7 @@ return 1
     """
 
     def __init__(self, member_id, parsed_url, options):
-        super(RedisDriver, self).__init__()
+        super(RedisDriver, self).__init__(member_id)
         options = utils.collapse(options, exclude=self.CLIENT_LIST_ARGS)
         self._parsed_url = parsed_url
         self._options = options
@@ -333,7 +339,6 @@ return 1
         self._beat_prefix = self._namespace + b"_beats"
         self._groups = self._namespace + b"_groups"
         self._client = None
-        self._member_id = utils.to_binary(member_id, encoding=self._encoding)
         self._acquired_locks = set()
         self._executor = utils.ProxyExecutor.build("Redis", options)
         self._started = False
@@ -427,9 +432,9 @@ return 1
             self._client = self._make_client(self._parsed_url, self._options,
                                              self.timeout)
         except exceptions.RedisError as e:
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.ToozConnectionError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         else:
             # Ensure that the server is alive and not dead, this does not
             # ensure the server will always be alive, but does insure that it
@@ -503,7 +508,7 @@ return 1
         for lock in self._acquired_locks.copy():
             try:
                 lock.heartbeat()
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to heartbeat lock '%s'", lock,
                             exc_info=True)
         return min(self.lock_timeout, self.membership_timeout)
@@ -513,7 +518,7 @@ return 1
             lock = self._acquired_locks.pop()
             try:
                 lock.release()
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to release lock '%s'", lock, exc_info=True)
         self._executor.stop()
         if self._client is not None:
@@ -524,7 +529,7 @@ return 1
                 # exist in the first place, which is fine/expected/desired...
                 with _translate_failures():
                     self._client.delete(beat_id)
-            except coordination.ToozError:
+            except tooz.ToozError:
                 LOG.warning("Unable to delete heartbeat key '%s'", beat_id,
                             exc_info=True)
             self._client = None
@@ -534,14 +539,14 @@ return 1
 
     def _submit(self, cb, *args, **kwargs):
         if not self._started:
-            raise coordination.ToozError("Redis driver has not been started")
+            raise tooz.ToozError("Redis driver has not been started")
         return self._executor.submit(cb, *args, **kwargs)
 
     def _get_script(self, script_key):
         try:
             return self._scripts[script_key]
         except KeyError:
-            raise coordination.ToozError("Redis driver has not been started")
+            raise tooz.ToozError("Redis driver has not been started")
 
     def create_group(self, group_id):
         script = self._get_script('create_group')
@@ -698,13 +703,13 @@ return 1
             if result == -3:
                 raise coordination.GroupNotEmpty(group_id)
             if result == -4:
-                raise coordination.ToozError("Unable to remove '%s' key"
-                                             " from set located at '%s'"
-                                             % (args[0], keys[-1]))
+                raise tooz.ToozError("Unable to remove '%s' key"
+                                     " from set located at '%s'"
+                                     % (args[0], keys[-1]))
             if result != 1:
-                raise coordination.ToozError("Internal error, unable"
-                                             " to complete group '%s' removal"
-                                             % (group_id))
+                raise tooz.ToozError("Internal error, unable"
+                                     " to complete group '%s' removal"
+                                     % (group_id))
 
         return RedisFutureResult(self._submit(_delete_group, script))
 
@@ -722,32 +727,6 @@ return 1
 
         return RedisFutureResult(self._submit(_get_groups))
 
-    def _init_watch_group(self, group_id):
-        members = self.get_members(group_id)
-        self._group_members[group_id].update(members.get(timeout=None))
-
-    def watch_join_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(RedisDriver, self).watch_join_group(group_id, callback)
-
-    def unwatch_join_group(self, group_id, callback):
-        return super(RedisDriver, self).unwatch_join_group(group_id, callback)
-
-    def watch_leave_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(RedisDriver, self).watch_leave_group(group_id, callback)
-
-    def unwatch_leave_group(self, group_id, callback):
-        return super(RedisDriver, self).unwatch_leave_group(group_id, callback)
-
-    def watch_elected_as_leader(self, group_id, callback):
-        return super(RedisDriver, self).watch_elected_as_leader(
-            group_id, callback)
-
-    def unwatch_elected_as_leader(self, group_id, callback):
-        return super(RedisDriver, self).unwatch_elected_as_leader(
-            group_id, callback)
-
     def _get_leader_lock(self, group_id):
         name = self._encode_group_leader(group_id)
         return self.get_lock(name)
@@ -781,9 +760,9 @@ class RedisFutureResult(coordination.CoordAsyncResult):
             with _translate_failures():
                 return self._fut.result(timeout=timeout)
         except futures.TimeoutError as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     def done(self):
         return self._fut.done()
